package jabadot;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

/** The main "properties"-like class for configuring JabaDot.
 * All of its properties except "jabadot.dir" are loaded from
 * a properties file. This program has been configured as a "servlet"
 * - which it really is not -- just so it can get the value for
 * JABADOT_DIR from the servlet parameters. As a result, this
 * file MUST be listed as "preload" in the servlet engine configuration.
 * AND FURTHERMORE, you must set the equivalent of "reloadable=false"
 * for this Web Application, since reloading trashes the classloader,
 * which trashes all static data, which trashes how this puppy works!
 *
 * @version $Id: JDConstants.java,v 1.13 2001/04/09 15:32:21 ian Exp $
 */
public class JDConstants extends HttpServlet {

	/** This must be a full path, since ya gotta start someplace. */
	protected static String JABADOT_DIR;

	/** The properties file within the JABADOT_DIR directory */
	protected final static String PROPERTIES = "jabadot.properties";

	protected static String propsFileFullPath;

	protected static Properties p;

	/** Static initialiser, just for monitoring. */
	static {
		System.out.println("JDConstants loaded!");
	}

	/** This init servlet method loads the Properties. We know we'll
	 * get called once at the beginning of time, since we're set up
	 * that way in web.xml (as a pre-loaded servlet).
	 */
	public void init() throws ServletException {

		ServletConfig cfg = getServletConfig();
		ServletContext ctx = getServletContext();

		// Get the JABADOT_DIR before all else!

		// Be careful of the difference between 
		// ServletContext.getInitParameter and Servlet.getInitParameter!

		JABADOT_DIR = ctx.getInitParameter("jabadot.dir");
		if (JABADOT_DIR == null) {
			JABADOT_DIR = getServletContext().getRealPath("/");
			System.out.println("JABADOT_DIR computed as " + JABADOT_DIR);
		} else {
			System.out.println("JABADOT_DIR set to " + JABADOT_DIR);
		}

		// Now save jabadot.dir in the Application scope, so that if
		// we get a RELOAD and lose everything (yes, even our
		// static final data), we can rebuild everything.
		// A tiny bit like surviving the GNAB GIB at the end of the universe...
		// PageContext ptx = request.getPageContext();
		// ptx.setAttribute("jabadot.dir", JABADOT_DIR,
		// 	PageContext.APPLICATION_SCOPE);
		// Not a JSP: best we can do?
		ctx.setAttribute("jabadot.dir", JABADOT_DIR);
	
		reloadProps();
	}

	/** Load the properties, assuming that JABADOT_DIR has been set */
	protected void reloadProps() {

		// This must be next
		propsFileFullPath = JABADOT_DIR + '/' + PROPERTIES;

		p = new Properties();
		p.setProperty("jabadot.dir", JABADOT_DIR);
		try {
			System.out.println("Opening Props File " + propsFileFullPath);
			InputStream is = new FileInputStream(propsFileFullPath);
			load(is);	// see below; includes substitution.
		} catch (Exception e) {
			System.err.println("Unable to load JabaDot Properties!! " + e);
		}
	}

	/** the service() method is invoked periodically to ensure that
	 * the JDConstants properties still work (they must get lost on
	 * a "RELOAD!!" since their classloader goes away...).
	 * Note: it doesn't have to DO *anything*, chuckle: the
	 * container will ensure that its init() gets called before this,
	 * after a RELOAD.
	 */
	public void service(HttpServletRequest req, HttpServletResponse resp) 
	throws ServletException, IOException {
		resp.setContentType("text/html");
		// System.err.println("in service, p = " + p);
		if (p==null) {
			JABADOT_DIR =
				(String) getServletContext().getAttribute("jabadot.dir");
			System.out.println("JABADOT_DIR resurrected as " + JABADOT_DIR);
			reloadProps();
		}
	}

	public static String getProperty(String pname) {
		if (p==null)
			throw new IllegalStateException(
				"JDConstants.getProperty: not initialized!");
		return p.getProperty(pname);
	}

	public static Properties getProperties() {
		return p;
	}

	/** Load a Properties, but substitute $DIR with JABADOT_DIR */
	private static void load(InputStream is) throws IOException {
		Properties tmpP = new Properties();
		tmpP.load(is);
		Enumeration en = tmpP.keys();
		while (en.hasMoreElements()) {
			String key = (String)en.nextElement();
			String val = tmpP.getProperty(key);
			String nval = replace("$DIR", JABADOT_DIR, val);
			p.setProperty(key, nval);
		}
	}

	/** Replace one string in another
	 * @return The modified string, or the original if no change made.
	 */
	public static String replace(String oldStr, String newStr, String inString) {
		int start = inString.indexOf(oldStr);
		if (start == -1) {
			return inString;
		}
		StringBuffer sb = new StringBuffer();
		sb.append(inString.substring(0, start));
		sb.append(newStr);
		sb.append(inString.substring(start+oldStr.length()));
		return sb.toString();
	}

	/** Simple main for testing purposes. */
	public static void main(String[] argv) {
		JDConstants.getProperties().list(System.out);
	}
}
